/* IRCQueue.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#include "IRCQueue.h"
#include "ClientSocket.h"
#include "IOHandler.h"
#include "tools.h"

#define MAXPENALTY 8 /* 4 messages */

struct QueueEntry {
    char *msg;
    struct QueueEntry *next;
};

struct BotQueue {
    struct ClientSocket *client;
    struct IODescriptor *iofd;
    int penalty : 8;
    int rem_penalty : 8;
    struct QueueEntry *fastqueue_first, *fastqueue_last;
    struct QueueEntry *normalqueue_first, *normalqueue_last;
    struct QueueEntry *textqueue_first, *textqueue_last;
};

static IOHANDLER_CALLBACK(queue_callback);

static struct BotQueue *initialize_queue(struct ClientSocket *client) {
    struct BotQueue *queue = malloc(sizeof(*queue));
    if (!queue) {
        perror("malloc() failed");
        return NULL;
    }
    queue->client = client;
    client->queue = queue;
    queue->iofd = NULL;
    queue->penalty = 0;
    queue->fastqueue_first = NULL;
    queue->fastqueue_last = NULL;
    queue->normalqueue_first = NULL;
    queue->normalqueue_last = NULL;
    queue->textqueue_first = NULL;
    queue->textqueue_last = NULL;
    return queue;
}

static int calculate_penalty(char *message) {
    int msglen = strlen(message);
    int penalty = (2 + msglen / 100);
    return penalty;
}

int queue_add(struct ClientSocket *client, char* msg, int len) {
    if(!client->queue)
        client->queue = initialize_queue(client);
    struct BotQueue *queue = client->queue;
    char *args = strstr(msg, " ");
    int type;
    int add_queue = 0;
    if(args) {
        *args = '\0';
        if(!stricmp(msg, "MODE")) 
            type = 3;
        else if(!stricmp(msg, "KICK")) 
            type = 3;
        else if(!stricmp(msg, "PONG")) 
            type = 3;
        else if(!stricmp(msg, "PRIVMSG")) 
            type = 1;
        else if(!stricmp(msg, "NOTICE")) 
            type = 1;
        else if(!stricmp(msg, "WHO")) 
            type = 1;
        else
            type = 2;
        *args = ' ';
    } else
        type = 2;
    
    //check if we need to queue
    switch(type) {
    case 1:
        if(queue->textqueue_first) {
            add_queue = 1;
            break;
        }
    case 2:
        if(queue->normalqueue_first) {
            add_queue = 1;
            break;
        }
    case 3:
        if(queue->fastqueue_first) {
            add_queue = 1;
            break;
        }
    default:
        if(queue->penalty >= MAXPENALTY)
            add_queue = 1;
        break;
    }
    
    if(!add_queue) {
        int penalty = calculate_penalty(msg);
        write_socket_force(client, msg, len);
        queue->penalty += penalty;
        if(!queue->iofd) {
            struct timeval timeout;
            gettimeofday(&timeout, NULL);
            if(queue->penalty >= MAXPENALTY)
                queue->rem_penalty = (queue->penalty - MAXPENALTY)+1;
            else
                queue->rem_penalty = queue->penalty;
            timeout.tv_sec += queue->rem_penalty;
            queue->iofd = iohandler_timer(timeout, queue_callback);
            queue->iofd->data = queue;
        }
    } else {
        struct QueueEntry *entry = malloc(sizeof(*entry));
        if (!entry) {
            perror("malloc() failed");
            return 0;
        }
        entry->msg = strdup(msg);
        entry->next = NULL;
        if(type == 1) { //low priority
            if(queue->textqueue_last) {
                queue->textqueue_last->next = entry;
                queue->textqueue_last = entry;
            } else {
                queue->textqueue_last = entry;
                queue->textqueue_first = entry;
            }
        } else if(type == 2) { //normal priority
            if(queue->normalqueue_last) {
                queue->normalqueue_last->next = entry;
                queue->normalqueue_last = entry;
            } else {
                queue->normalqueue_last = entry;
                queue->normalqueue_first = entry;
            }
        } else if(type == 3) { //high priority
            if(queue->fastqueue_last) {
                queue->fastqueue_last->next = entry;
                queue->fastqueue_last = entry;
            } else {
                queue->fastqueue_last = entry;
                queue->fastqueue_first = entry;
            }
        }
    }
    return 1;
}

static void dequeue_bot(struct ClientSocket *client) {
    if(client->queue->penalty >= MAXPENALTY) return;
    int penalty;
    //try to send high priority messages
    if(client->queue->fastqueue_first) {
        do {
            struct QueueEntry *entry = client->queue->fastqueue_first;
            if(!entry->next)
                client->queue->fastqueue_last = NULL;
            client->queue->fastqueue_first = client->queue->fastqueue_first->next;
            penalty = calculate_penalty(entry->msg);
            write_socket_force(client, entry->msg, strlen(entry->msg));
            client->queue->penalty += penalty;
            free(entry->msg);
            free(entry);
        } while(client->queue->penalty < MAXPENALTY && client->queue->fastqueue_first);
    }
    if(client->queue->penalty >= MAXPENALTY) return;
    //try to send normal priority messages
    if(client->queue->normalqueue_first) {
        do {
            struct QueueEntry *entry = client->queue->normalqueue_first;
            if(!entry->next)
                client->queue->normalqueue_last = NULL;
            client->queue->normalqueue_first = client->queue->normalqueue_first->next;
            penalty = calculate_penalty(entry->msg);
            write_socket_force(client, entry->msg, strlen(entry->msg));
            client->queue->penalty += penalty;
            free(entry->msg);
            free(entry);
        } while(client->queue->penalty < MAXPENALTY && client->queue->normalqueue_first);
    }
    if(client->queue->penalty >= MAXPENALTY) return;
    //try to send low priority messages
    if(client->queue->textqueue_first) {
        do {
            struct QueueEntry *entry = client->queue->textqueue_first;
            if(!entry->next)
                client->queue->textqueue_last = NULL;
            client->queue->textqueue_first = client->queue->textqueue_first->next;
            penalty = calculate_penalty(entry->msg);
            write_socket_force(client, entry->msg, strlen(entry->msg));
            client->queue->penalty += penalty;
            free(entry->msg);
            free(entry);
        } while(client->queue->penalty < MAXPENALTY && client->queue->textqueue_first);
    }
}

void queue_destroy(struct ClientSocket *client) {
    if(!client->queue) return;
    struct QueueEntry *entry, *next;
    for(entry = client->queue->fastqueue_first; entry; entry = next) {
        next = entry->next;
        free(entry->msg);
        free(entry);
    }
    for(entry = client->queue->normalqueue_first; entry; entry = next) {
        next = entry->next;
        free(entry->msg);
        free(entry);
    }
    for(entry = client->queue->textqueue_first; entry; entry = next) {
        next = entry->next;
        free(entry->msg);
        free(entry);
    }
    if(client->queue->iofd)
        iohandler_close(client->queue->iofd);
    free(client->queue);
    client->queue = NULL;
}

static IOHANDLER_CALLBACK(queue_callback) {
    struct BotQueue *queue = event->iofd->data;
    switch(event->type) {
    case IOEVENT_TIMEOUT:
        queue->penalty -= queue->rem_penalty;
        dequeue_bot(queue->client);
        if(queue->penalty > 0) {
            struct timeval timeout;
            gettimeofday(&timeout, NULL);
            if(queue->penalty >= MAXPENALTY)
                queue->rem_penalty = (queue->penalty - MAXPENALTY)+1;
            else
                queue->rem_penalty = queue->penalty;
            timeout.tv_sec += queue->rem_penalty;
            queue->iofd = iohandler_timer(timeout, queue_callback);
            queue->iofd->data = queue;
        } else {
            queue->iofd = NULL;
            queue->penalty = 0;
        }
        break;
    default:
        break;
    }
}
