/* lang.c - NeonServ v5.6
 * Copyright (C) 2011-2012  Philipp Kreil (pk910)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#include "lang.h"
#include "UserNode.h"
#include "DBHelper.h"
#include "mysqlConn.h"
#include "tools.h"

#define DEFAULT_LANG_TAG "EN"
#define DEFAULT_LANG_NAME "English"

static struct language **langdict;
static struct language *lang_c;

void init_lang() {
    langdict = calloc(MAXLANGUAGES, sizeof(*langdict));
}

void free_lang() {
    
}

void load_languages() {
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `lang`, `text` FROM `language` WHERE `ident` = 'name'");
    res = mysql_use();
    while((row = mysql_fetch_row(res)) != NULL) {
        load_language(row[0], row[1]);
    }
}

static struct language* add_language(char *langtag, char *langname) {
    int cindex;
    for(cindex = 0; cindex < MAXLANGUAGES; cindex++) {
        if(langdict[cindex] == NULL) break;
        if(!strcmp(langdict[cindex]->langname, langname) || !strcmp(langdict[cindex]->langtag, langtag))
            return langdict[cindex];
    }
    if(cindex == MAXLANGUAGES) return NULL;
    struct language *lang = malloc(sizeof(*lang));
    if (!lang) {
        perror("malloc() failed");
        return NULL;
    }
    lang->langtag = strdup(langtag);
    lang->langname = strdup(langname);
    struct language_table **entrys = calloc(27, sizeof(*entrys));
    lang->entrys = entrys;
    langdict[cindex] = lang;
    return lang;
}

static int get_entry_index(const char *ident) {
    const char *underscore = strstr(ident, "_");
    if(!underscore || !(underscore[1] >= 65 && underscore[1] <= 90)) return 26;
    return (underscore[1] - 'A');
}

void load_language(char *tag, char *name) {
    struct language *lang = get_language_by_tag(tag);
    if(lang == get_default_language()) return;
    if(lang) {
        //remove all entrys
        int cindex;
        struct language_table *entry, *next;
        for(cindex = 0; cindex < 27; cindex++) {
            for(entry = lang->entrys[cindex]; entry; entry = next) {
                next = entry->next;
                free(entry->ident);
                free(entry->text);
                free(entry);
            }
            lang->entrys[cindex] = NULL;
        }
    } else
        lang = add_language(tag, name);
    if(!lang) return;
    MYSQL_RES *res;
    MYSQL_ROW row;
    printf_mysql_query("SELECT `ident`, `text` FROM `language` WHERE `lang` = '%s' AND `ident` != 'name'", escape_string(tag));
    res = mysql_use();
    while((row = mysql_fetch_row(res)) != NULL) {
        register_language_string(lang, row[0], row[1]);
    }
}

struct language* get_language_by_tag(char *tag) {
    int cindex;
    for(cindex = 0; cindex < MAXLANGUAGES; cindex++) {
        if(langdict[cindex] == NULL) break;
        if(!stricmp(langdict[cindex]->langtag, tag))
            return langdict[cindex];
    }
    return NULL;
}

struct language* get_language_by_name(char *name) {
    int cindex;
    for(cindex = 0; cindex < MAXLANGUAGES; cindex++) {
        if(langdict[cindex] == NULL) break;
        if(!stricmp(langdict[cindex]->langname, name))
            return langdict[cindex];
    }
    return NULL;
}

struct language* get_default_language() {
    if(lang_c == NULL) 
        lang_c = add_language(DEFAULT_LANG_TAG, DEFAULT_LANG_NAME);
    return lang_c;
}

void register_default_language_table(const struct default_language_entry *msgtab) {
    if(lang_c == NULL) 
        lang_c = add_language(DEFAULT_LANG_TAG, DEFAULT_LANG_NAME);
    while(msgtab->ident) {
        register_language_string(lang_c, msgtab->ident, msgtab->text);
        msgtab++;
    }
}

void register_language_string(struct language *lang, char *ident, char *text) {
    int cindex = get_entry_index(ident);
    struct language_table *lang_entry;
    for(lang_entry = lang->entrys[cindex]; lang_entry; lang_entry = lang_entry->next) {
        if(!strcmp(lang_entry->ident, ident)) break;
    }
    if(!lang_entry) {
        lang_entry = malloc(sizeof(*lang_entry));
        if (!lang_entry) {
            perror("malloc() failed");
            return;
        }
        lang_entry->ident = strdup(ident);
        lang_entry->next = lang->entrys[cindex];
        lang->entrys[cindex] = lang_entry;
    } else
        free(lang_entry->text); //free old text (new one will be set below)
    //replace all:
    //$b to \002
    //$k to \003
    char txt[MAXLEN];
    strcpy(txt, text);
    char tmp[MAXLEN];
    int tmppos = 0;
    char *a, *b = txt;
    do {
        a = strstr(b, "$");
        if(a) *a = '\0';
        tmppos += sprintf(tmp + tmppos, "%s", b);
        if(a) {
            switch(a[1]) {
                case 'b':
                    tmp[tmppos++] = 2;
                    break;
                case 'k':
                    tmp[tmppos++] = 3;
                    break;
                case 'u':
                    tmp[tmppos++] = 31;
                    break;
                default:
                    //unknown - just write it
                    tmppos += sprintf(tmp + tmppos, "$%c", a[1]);
            }
            b = a+2;
        }
    } while(a);
    lang_entry->text = strdup(tmp);
}

char *get_language_string(struct UserNode *user, const char* msg_ident) {
    struct language* lang;
    if(user && (user->flags & USERFLAG_ISAUTHED)) {
        loadUserSettings(user);
        lang = user->language;
    } else
        lang = lang_c;
    int cindex = get_entry_index(msg_ident);
    struct language_table* entry;
    for(entry = lang->entrys[cindex]; entry; entry = entry->next) {
        if(!strcmp(entry->ident, msg_ident))
            return entry->text;
    }
    if(lang == lang_c) return NULL;
    for(entry = lang_c->entrys[cindex]; entry; entry = entry->next) {
        if(!strcmp(entry->ident, msg_ident))
            return entry->text;
    }
    return NULL;
}

char *build_language_string(struct UserNode *user, char *buffer, const char *msg_ident, ...) {
    char *formatStr = get_language_string(user, msg_ident);
    if(!formatStr) return NULL;
    if(buffer == NULL) {
        buffer = (char *)malloc((MAXLEN+1) * sizeof(char));
        if (!buffer) {
            perror("malloc() failed");
            return NULL;
        }
    }
    int pos;
    va_list arg_list;
    buffer[0] = '\0';
    va_start(arg_list, msg_ident);
    pos = vsnprintf(buffer, MAXLEN - 2, formatStr, arg_list);
    va_end(arg_list);
    if (pos < 0 || pos > (MAXLEN - 2)) pos = MAXLEN - 2;
    buffer[pos] = '\0';
    return buffer;
}
